@@ -1,5 +1,5 @@
 /*
- * Gophernicus - Copyright (c) 2009-2010 Kim Holviala <kim@holviala.com>
+ * Gophernicus - Copyright (c) 2009-2014 Kim Holviala <kim@holviala.com>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -26,18 +26,6 @@
 #include "gophernicus.h"
 
 
-/*
- * Return bigger of two integers
- *
- * Hmmm... Why is this a function and not a #define? *wondering*
- */
-inline int max(int a, int b)
-{
-	if (a > b) return a;
-	return b;
-}
-
-
 /*
  * Print gopher menu line
  */
@@ -126,10 +114,7 @@ void die(state *st, char *message, char *description)
 
 	/* Handle menu errors */
 	if (st->req_filetype == TYPE_MENU || st->req_filetype == TYPE_QUERY) {
-#ifdef ENABLE_STRICT_RFC1436
 		printf("3" ERROR_PREFIX "%s\tTITLE\t" DUMMY_HOST CRLF, message);
-#endif
-		printf("i" ERROR_PREFIX "%s\tTITLE\t" DUMMY_HOST CRLF, message);
 		footer(st);
 	}
 
@@ -206,10 +191,33 @@ void selector_to_path(state *st)
 	struct stat file;
 #ifdef HAVE_PASSWD
 	struct passwd *pwd;
-	char buf[BUFSIZE];
 	char *path = EMPTY;
 	char *c;
+#endif
+	char buf[BUFSIZE];
+	int i;
 
+	/* Handle selector rewriting */
+	for (i = 0; i < st->rewrite_count; i++) {
+
+		/* Match found? */
+		if (strstr(st->req_selector, st->rewrite[i].match) == st->req_selector) {
+
+			/* Replace match with a new string */
+			snprintf(buf, sizeof(buf), "%s%s",
+				st->rewrite[i].replace,
+				st->req_selector + strlen(st->rewrite[i].match));
+
+			if (st->debug) {
+				syslog(LOG_INFO, "rewriting selector \"%s\" -> \"%s\"",
+					st->req_selector, buf);
+			}
+
+			sstrlcpy(st->req_selector, buf);
+		}
+	}
+
+#ifdef HAVE_PASSWD
 	/* Virtual userdir (~user -> /home/user/public_gopher)? */
 	if (*(st->user_dir) && sstrncmp(st->req_selector, "/~") == MATCH) {
 
@@ -284,6 +292,47 @@ void selector_to_path(state *st)
 }
 
 
+/*
+ * Get local IP address
+ */
+char *get_local_address(void)
+{
+#ifdef HAVE_IPv4
+	struct sockaddr_in addr;
+	socklen_t addrsize = sizeof(addr);
+#endif
+#ifdef HAVE_IPv6
+	struct sockaddr_in6 addr6;
+	socklen_t addr6size = sizeof(addr6);
+	static char address[INET6_ADDRSTRLEN];
+#endif
+	char *c;
+
+	/* Try IPv4 first */
+#ifdef HAVE_IPv4
+	if (getsockname(0, (struct sockaddr *) &addr, &addrsize) == OK) {
+		c = inet_ntoa(addr.sin_addr);
+		if (strlen(c) > 0 && *c != '0') return c;
+	}
+#endif
+
+	/* IPv4 didn't work - try IPv6 */
+#ifdef HAVE_IPv6
+	if (getsockname(0, (struct sockaddr *) &addr6, &addr6size) == OK) {
+		if (inet_ntop(AF_INET6, &addr6.sin6_addr, address, sizeof(address))) {
+
+			/* Strip ::ffff: IPv4-in-IPv6 prefix */
+			if (sstrncmp(address, "::ffff:") == MATCH) return (address + 7);
+			else return address;
+		}
+	}
+#endif
+
+	/* Nothing works... I'm out of ideas */
+	return DEFAULT_ADDR;
+}
+
+
 /*
  * Get remote peer IP address
  */
@@ -335,6 +384,8 @@ char *get_peer_address(void)
 void init_state(state *st)
 {
 	static const char *filetypes[] = { FILETYPES };
+	char buf[BUFSIZE];
+	char *c;
 	int i;
 
 	/* Request */
@@ -342,20 +393,26 @@ void init_state(state *st)
 	strclear(st->req_realpath);
 	strclear(st->req_query_string);
 	strclear(st->req_referrer);
-	sstrlcpy(st->req_remote_addr, DEFAULT_ADDR);
+	sstrlcpy(st->req_local_addr, get_local_address());
+	sstrlcpy(st->req_remote_addr, get_peer_address());
 	/* strclear(st->req_remote_host); */
 	st->req_filetype = DEFAULT_TYPE;
+	st->req_protocol = PROTO_GOPHER;
 	st->req_filesize = 0;
 
 	/* Output */
 	st->out_width = DEFAULT_WIDTH;
-	sstrlcpy(st->out_charset, DEFAULT_CHARSET);
+	st->out_charset = DEFAULT_CHARSET;
 
 	/* Settings */
-	strclear(st->server_platform);
 	sstrlcpy(st->server_root, DEFAULT_ROOT);
 	sstrlcpy(st->server_host_default, DEFAULT_HOST);
-	sstrlcpy(st->server_host, DEFAULT_HOST);
+
+	if ((c = getenv("HOSTNAME")))
+		sstrlcpy(st->server_host, c);
+	else if ((gethostname(buf, sizeof(buf))) != ERROR)
+		sstrlcpy(st->server_host, buf);
+
 	st->server_port = DEFAULT_PORT;
 
 	st->default_filetype = DEFAULT_TYPE;
@@ -368,9 +425,12 @@ void init_state(state *st)
 	st->hidden_count = 0;
 	st->filetype_count = 0;
 	strclear(st->filter_dir);
+	st->rewrite_count = 0;
 
 	strclear(st->server_description);
-	sstrlcpy(st->gopher_proxy, DEFAULT_PROXY);
+	strclear(st->server_location);
+	strclear(st->server_platform);
+	strclear(st->server_admin);
 
 	/* Session */
 	st->session_timeout = DEFAULT_SESSION_TIMEOUT;
@@ -387,7 +447,9 @@ void init_state(state *st)
 	st->opt_magic = TRUE;
 	st->opt_iconv = TRUE;
 	st->opt_query = TRUE;
+	st->opt_caps = TRUE;
 	st->opt_shm = TRUE;
+	st->opt_root = TRUE;
 	st->debug = FALSE;
 
 	/* Load default suffix -> filetype mappings */
@@ -401,127 +463,6 @@ void init_state(state *st)
 }
 
 
-/*
- * Add one suffix->filetype mapping to the filetypes array
- */
-void add_ftype_mapping(state *st, char *suffix)
-{
-	char *type;
-	int i;
-
-	/* Let's not do anything stupid */
-	if (!*suffix) return;
-	if (!(type = strchr(suffix, '='))) return;
-
-	/* Extract type from the suffix:X string */
-	*type++ = '\0';
-	if (!*type) return;
-
-	/* Loop through the filetype array */
-	for (i = 0; i < st->filetype_count; i++) {
-
-		/* Old entry found? */
-		if (strcasecmp(st->filetype[i].suffix, suffix) == MATCH) {
-			st->filetype[i].type = *type;
-			return;
-		}
-	}
-
-	/* No old entry found - add new entry */
-	if (i < MAX_FILETYPES) {
-		sstrlcpy(st->filetype[i].suffix, suffix);
-		st->filetype[i].type = *type;
-		st->filetype_count++;
-	}
-}
-
-
-/*
- * Parse command-line arguments
- */
-void parse_args(state *st, int argc, char *argv[])
-{
-	FILE *fp;
-	static const char readme[] = README;
-	static const char license[] = LICENSE;
-	struct stat file;
-	char buf[BUFSIZE];
-	int opt;
-
-	/* Parse args */
-	while ((opt = getopt(argc, argv, "h:p:r:t:g:a:c:u:m:w:o:s:i:k:f:e:D:L:P:n:dl?-")) != ERROR) {
-		switch(opt) {
-			case 'h': sstrlcpy(st->server_host, optarg); break;
-			case 'p': st->server_port = atoi(optarg); break;
-			case 'r': sstrlcpy(st->server_root, optarg); break;
-			case 't': st->default_filetype = *optarg; break;
-			case 'g': sstrlcpy(st->map_file, optarg); break;
-			case 'a': sstrlcpy(st->map_file, optarg); break;
-			case 'c': sstrlcpy(st->cgi_file, optarg); break;
-			case 'u': sstrlcpy(st->user_dir, optarg);  break;
-			case 'm': sstrlcpy(st->log_file, optarg);  break;
-
-			case 'w': st->out_width = atoi(optarg); break;
-			case 'o': sstrlcpy(st->out_charset, optarg);  break;
-
-			case 's': st->session_timeout = atoi(optarg); break;
-			case 'i': st->session_max_kbytes = abs(atoi(optarg)); break;
-			case 'k': st->session_max_hits = abs(atoi(optarg)); break;
-
-			case 'f': sstrlcpy(st->filter_dir, optarg); break;
-			case 'e': add_ftype_mapping(st, optarg); break;
-
-			case 'D': sstrlcpy(st->server_description, optarg); break;
-			case 'P': sstrlcpy(st->gopher_proxy, optarg); break;
-
-			case 'n':
-				if (*optarg == 'v') { st->opt_vhost = FALSE; break; }
-				if (*optarg == 'l') { st->opt_parent = FALSE; break; }
-				if (*optarg == 'h') { st->opt_header = FALSE; break; }
-				if (*optarg == 'f') { st->opt_footer = FALSE; break; }
-				if (*optarg == 'd') { st->opt_date = FALSE; break; }
-				if (*optarg == 'c') { st->opt_magic = FALSE; break; }
-				if (*optarg == 'o') { st->opt_iconv = FALSE; break; }
-				if (*optarg == 'q') { st->opt_query = FALSE; break; }
-				if (*optarg == 's') { st->opt_syslog = FALSE; break; }
-				if (*optarg == 'm') { st->opt_shm = FALSE; break; }
-				break;
-
-			case 'd': st->debug = TRUE; break;
-			case 'l': printf(license); exit(EXIT_SUCCESS);
-			default : printf(readme); exit(EXIT_SUCCESS);
-		}
-	}
-
-	/* Sanitize options */
-	if (st->out_width > MAX_WIDTH) st->out_width = MAX_WIDTH;
-	if (st->out_width < MIN_WIDTH) st->out_width = MIN_WIDTH;
-	if (st->out_width  < MIN_WIDTH + DATE_WIDTH) st->opt_date = FALSE;
-	if (!st->opt_syslog) st->debug = FALSE;
-
-	/* Primary vhost directory must exist or we disable vhosting */
-	if (st->opt_vhost) {
-		snprintf(buf, sizeof(buf), "%s/%s", st->server_root, st->server_host);
-		if (stat(buf, &file) == ERROR) {
-			st->opt_vhost = FALSE;
-			if (st->debug)
-				syslog(LOG_INFO, "disabling vhosting: %s must exist", buf);
-		}
-	}
-
-	/* If -D arg looks like a file load the file contents */
-	if (*st->server_description == '/') {
-
-		if ((fp = fopen(st->server_description , "r"))) {
-			fgets(st->server_description, sizeof(st->server_description), fp);
-			chomp(st->server_description);
-			fclose(fp);
-		}
-		else strclear(st->server_description);
-	}
-}
-
-
 /*
  * Main
  */
@@ -530,10 +471,10 @@ int main(int argc, char *argv[])
 	struct stat file;
 	state st;
 	char self[64];
+	char selector[BUFSIZE];
 	char buf[BUFSIZE];
 	char *dest;
 	char *c;
-	int i;
 #ifdef HAVE_SHMEM
 	struct shmid_ds shm_ds;
 	shm_state *shm;
@@ -549,21 +490,26 @@ int main(int argc, char *argv[])
 	setlocale(LC_TIME, DATE_LOCALE);
 #endif
 	init_state(&st);
-
-	/* Refuse to run as root */
-	if (getuid() == 0)
-		die(&st, ERR_ACCESS, "Refusing to run as root");
-
-	/* Get server hostname */
-	if ((c = getenv("HOSTNAME"))) sstrlcpy(st.server_host, c);
-	else if ((gethostname(buf, sizeof(buf))) != ERROR) sstrlcpy(st.server_host, buf);
-
-	/* Get remote peer IP */
-	sstrlcpy(st.req_remote_addr, get_peer_address());
+	srand(time(NULL) / (getpid() + getppid()));
 
 	/* Handle command line arguments */
 	parse_args(&st, argc, argv);
 
+	/* Open syslog() */
+	if (st.opt_syslog) openlog(self, LOG_PID, LOG_DAEMON);
+
+	/* Make sure the computer is turned on */
+#ifdef __HAIKU__
+	if (is_computer_on() != TRUE)
+		die(&st, ERR_ACCESS, "Please turn on the computer first");
+#endif
+
+	/* Refuse to run as root */
+#ifdef HAVE_PASSWD
+	if (st.opt_root && getuid() == 0)
+		die(&st, ERR_ACCESS, "Refusing to run as root");
+#endif
+
 	/* Try to get shared memory */
 #ifdef HAVE_SHMEM
 	if ((shmid = shmget(SHM_KEY, sizeof(shm_state), IPC_CREAT | SHM_MODE)) == ERROR) {
@@ -590,10 +536,8 @@ int main(int argc, char *argv[])
 
 	/* For debugging shared memory issues */
 	if (!st.opt_shm) shm = NULL;
-#endif
 
 	/* Get server platform and description */
-#ifdef HAVE_SHMEM
 	if (shm) {
 		sstrlcpy(st.server_platform, shm->server_platform);
 
@@ -604,54 +548,48 @@ int main(int argc, char *argv[])
 #endif
 		platform(&st);
 
-	/* Open syslog() */
-	if (st.opt_syslog) openlog(self, LOG_PID, LOG_DAEMON);
+	/* Read selector */
+	if (fgets(selector, sizeof(selector) - 1, stdin) == NULL)
+		selector[0] = '\0';
 
-	/* Read selector, remove CRLF & encodings */
-	buf[0] = '/';
-	if (fgets(buf + 1, sizeof(buf) - 2, stdin) == NULL) buf[1] = '\0';
+	/* Remove trailing CRLF */
+	chomp(selector);
 
-	chomp(buf);
-	strndecode(buf, buf, sizeof(buf));	/* Do decoding in-place :) */
-
-	if (st.debug) syslog(LOG_INFO, "client sent us \"%s\"", buf + 1);
+	if (st.debug) syslog(LOG_INFO, "client sent us \"%s\"", selector);
 
 	/* Handle hURL: redirect page */
-	if (sstrncmp(buf + 1, "URL:") == MATCH) {
+	if (sstrncmp(selector, "URL:") == MATCH) {
 		st.req_filetype = TYPE_HTML;
-		sstrlcpy(st.req_selector, buf + 1);
+		sstrlcpy(st.req_selector, selector);
 		url_redirect(&st);
 		return OK;
 	}
 
-	/* Handle HTTP /server-status requests */
-#ifdef HAVE_SHMEM
-	if (sstrncmp(buf + 1, "GET " SERVER_STATUS) == MATCH) {
-		printf("HTTP/1.0 200 OK" CRLF);
-		printf("Content-Type: text/plain" CRLF);
-		printf("Server: " SERVER_SOFTWARE_FULL CRLF CRLF, st.server_platform);
-
-		/* Make logging accurate */
-		sstrlcpy(st.req_selector, SERVER_STATUS);
-
-		if (shm) server_status(&st, shm, shmid);
-		return OK;
-	}
-#endif
-
-	/* Redirect HTTP requests to a gopher proxy */
-	if (sstrncmp(buf + 1, "GET ") == MATCH) {
-		if (st.debug) syslog(LOG_INFO, "got http request, redirecting to proxy");
-
-		printf("HTTP/1.0 301 Moved Permanently" CRLF);
-		printf("Location: %s%s:%i/" CRLF,
-			st.gopher_proxy,
+	/* Handle gopher+ root requests (UMN gopher client is seriously borken) */
+	if (sstrncmp(selector, "\t$") == MATCH) {
+		printf("+-1" CRLF);
+		printf("+INFO: 1Main menu\t\t%s\t%i" CRLF,
 			st.server_host,
 			st.server_port);
-		printf("Server: " SERVER_SOFTWARE_FULL CRLF CRLF, st.server_platform);
+		printf("+VIEWS:" CRLF " application/gopher+-menu: <512b>" CRLF);
+		printf("." CRLF);
+
+		if (st.debug) syslog(LOG_INFO, "got a request for gopher+ root menu");
 		return OK;
 	}
 
+	/* Convert HTTP request to gopher (respond using headerless HTTP/0.9) */
+	if (sstrncmp(selector, "GET ") == MATCH ||
+	    sstrncmp(selector, "POST ") == MATCH ) {
+
+		if ((c = strchr(selector, ' '))) sstrlcpy(selector, c + 1);
+		if ((c = strchr(selector, ' '))) *c = '\0';
+
+		st.req_protocol = PROTO_HTTP;
+
+		if (st.debug) syslog(LOG_INFO, "got HTTP request for \"%s\"", selector);
+	}
+
 	/* Save default server_host & fetch session data (including new server_host) */
 	sstrlcpy(st.server_host_default, st.server_host);
 #ifdef HAVE_SHMEM
@@ -660,8 +598,9 @@ int main(int argc, char *argv[])
 
 	/* Loop through the selector, fix it & separate query_string */
 	dest = st.req_selector;
+	if (selector[0] != '/') *dest++ = '/';
 
-	for (c = buf; *c;) {
+	for (c = selector; *c;) {
 
 		/* Skip duplicate slashes and /./ */
 		while (*c == '/' && *(c + 1) == '/') c++;
@@ -688,7 +627,10 @@ int main(int argc, char *argv[])
 	}
 	*dest = '\0';
 
-	/* Deny Slashdot users and /../ hackers */
+	/* Remove encodings from selector */
+	strndecode(st.req_selector, st.req_selector, sizeof(st.req_selector));
+
+	/* Deny requests for Slashdot and /../ hackers */
 	if (strstr(st.req_selector, "/."))
 		die(&st, ERR_ACCESS, "Refusing to serve out dotfiles");
 
@@ -700,25 +642,33 @@ int main(int argc, char *argv[])
 	}
 #endif
 
-	/* Handle /caps.txt requests */
-	if (sstrncmp(st.req_selector, CAPS_TXT) == MATCH) {
-		caps_txt(&st, shm);
-		return OK;
-	}
-
 	/* Remove possible extra cruft from server_host */
 	if ((c = strchr(st.server_host, '\t'))) *c = '\0';
 
 	/* Guess request filetype so we can die() with style... */
-	i = st.opt_magic;
-	st.opt_magic = FALSE;	/* Don't stat() now */
-	st.req_filetype = gopher_filetype(&st, st.req_selector);
-	st.opt_magic = i;
+	st.req_filetype = gopher_filetype(&st, st.req_selector, FALSE);
 
 	/* Convert seletor to path & stat() */
 	selector_to_path(&st);
 	if (st.debug) syslog(LOG_INFO, "path to resource is \"%s\"", st.req_realpath);
-	if (stat(st.req_realpath, &file) == ERROR) die(&st, ERR_NOTFOUND, NULL);
+
+	if (stat(st.req_realpath, &file) == ERROR) {
+
+		/* Handle virtual /caps.txt requests */
+		if (st.opt_caps && sstrncmp(st.req_selector, CAPS_TXT) == MATCH) {
+#ifdef HAVE_SHMEM
+			caps_txt(&st, shm);
+#else
+			caps_txt(&st, NULL);
+#endif
+			return OK;
+		}
+
+		/* Requested file not found - die() */
+		die(&st, ERR_NOTFOUND, NULL);
+	}
+
+	/* Fetch request filesize from stat() */
 	st.req_filesize = file.st_size;
 
 	/* Everyone must have read access but no write access */
@@ -732,7 +682,7 @@ int main(int argc, char *argv[])
 
 	/* Not a dir - let's guess the filetype again... */
 	else if ((file.st_mode & S_IFMT) == S_IFREG)
-		st.req_filetype = gopher_filetype(&st, st.req_realpath);
+		st.req_filetype = gopher_filetype(&st, st.req_realpath, st.opt_magic);
 
 	/* Menu selectors must end with a slash */
 	if (st.req_filetype == TYPE_MENU && strlast(st.req_selector) != '/')
